/*------------------------------------------------------------------------*/
/*  File: comp_feat.cpp                                                   */
/*  Feature computation for general MIP instances, including all types    */
/*  of instances CPLEX can read.                                          */
/*------------------------------------------------------------------------*/
/*  This code started as a generalization of feature computation code     */
/*  for combinatorial auction instances written by Eugene Nudelman.       */
/*  Since then, we added lots of new features.                            */
/*  The CPLEX interfacing is based on file examples/src/lpex2.c in        */
/*  CPLEX, Version 9.1                                                    */
/*  Frank Hutter, Aug 19, 2008                                            */
/*------------------------------------------------------------------------*/
/*  Differences in features compared to Nudelman's code are due to        */
/*  normalizations that were done in that code (but not reported in       */
/*  papers), and to at least one bug in that code: there is a problem in  */
/*  the bid graph degrees in that the code counts hundreds of times too   */
/*  high degrees, resulting in edge densities > 500 (should be            */
/*  upper-bounded by 1). Also, in one run valgrind found 7373 errors from */
/*  121 contexts, which suggests there may be some problem with memory.   */
/*------------------------------------------------------------------------*/
/*  I am being sloppy with freeing memory - this                          */
/*  is just fine when used as a standalone application, but if you want   */
/*  to use it in a bigger system you'll have to fix it to prevent leaks.  */
/*------------------------------------------------------------------------*/

/* Bring in the CPLEX function declarations and the C library 
   header file stdio.h with the following single include. */

#include <ilcplex/cplex.h>

/* Bring in the declarations for the string and character functions 
   and malloc, timer, and set. */

#include <ctype.h>
#include <cmath>
#include <stdlib.h>
#include <string.h>
#include "stopwatch.h"
#include <set>
#include <vector>
#include <iostream>

#include "comp_feat.h"
#include <boost/lexical_cast.hpp>
using namespace std;

/* min and max macros */
#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))

/* Useful macro to swap 2 elements, using
 * the standard trick do { } while(0) to
 * make this one statement (otherwise it
 * could be wrong to use like 'else SWAP(A,B);').
 */
#define SWAP(A,B) \
do { int tmp = A; A = B; B = tmp; } while(0)

/* Transfer a string to double */
// TODO: rename
double str_to_f(string newstr){
	double myvalue ; 
	string tmp;
	istringstream in(newstr);
	double one=1.0;
	double two=2.0;
	double three=3.0;
	double four=4.0;
	// in >> tmp;

	if (newstr.find("*")!=string::npos){
		return one;
	}
	if (newstr.find("+")!=string::npos)
		return one;

	if (newstr.find("Cuts")!=string::npos)
		return one;

	if (newstr.find("Cliques")!=string::npos)
		return two;

	if (newstr.find("integral")!=string::npos)
		return one;

	in >> myvalue;
	return myvalue;
}

string convertInt(int number){
	return boost::lexical_cast<string>( number );
}

static void
   free_and_null (char **ptr),
   usage         (char *progname);


int size_of_intersection(int *a, int*b, int num_a, int num_b){
	/* Compute the size of the intersection of a[0],...,a[num_a-1], and b[0],...,b[num_b-1]
	   This assumes a and b are unique and sorted. */
	int i=0, j=0, jstart=0, count=0;
	while (i<num_a && j<num_b){
		if( a[i] > b[j] ){
			j++;
		} else if( a[i] < b[j] ){
			i++;
		} else {
			count++;
			i++;
			j++;
		}
	}
	return count;
}

void quick(int *a, int min, int max) {
	if (max - min > 1) {
		int i = min;
		int j = max;
		/* indices are positive */
		int pivot = a[(i+j) >> 1];
		do {
			while(a[i] < pivot) i++;
			while(a[j] > pivot) j--;
			if (i > j) break;
			SWAP(a[i], a[j]);
		} while(++i <= --j);

		/* Try to reduce bad behaviours. */
		while (min < j && a[j] == pivot) j--;
			if (min < j) quick(a, min, j);

		/* Try to reduce bad behaviours. */
		while (i < max && a[i] == pivot) i++;
			if (i < max) quick(a, i, max);

	} else if (a[min] > a[max])
		SWAP(a[min], a[max]);
}

/* Quick sort for sorting an int array a.
   Example: quick_sort(a, size);
// Equivalent: qsort(a, size, sizeof(int), compare_ints);
*/
void quick_sort(int *a, int size) {
	if (size > 1) {
		quick(a, 0, size-1);
	}
}

int compare_ints(const void * a, const void * b){
	const int *ia = (const int *) a;
	const int *ib = (const int *) b;
	return (*ia > *ib) - (*ia < *ib);
}


void printMatrixDouble(const double* X,int nRows, int nCols){
    int i,j;
    for(i = 0; i < nRows; i++) {
        printf("< ");
        for(j = 0; j < nCols; j++) {
            printf("%lf ",X[i+nRows*j]);
		}
        printf(">\n");
	}
}

void printMatrixInt(int* X,int nRows, int nCols){
    int i,j;
    for(i = 0; i < nRows; i++) {
        printf("< ");
        for(j = 0; j < nCols; j++) {
            printf("%d ",X[i+nRows*j]);
		}
        printf(">\n");
	}
}

/*
double i_var(const int *array, const int N){
    int i;
    int sum_x, sum_x_squared;
	if(N<=1) return 0;
    sum_x = 0;
    sum_x_squared = 0;
    for(i=0; i<N; i++){
        sum_x += array[i];
        sum_x_squared += array[i]*array[i];
    }
//    printMatrixDouble(array, N, 1);
//    printf("%lf, %lf, %d, %lf\n", sum_x, sum_x_squared, N, 1.0/(N-1) * (sum_x_squared - 1.0/N*sum_x*sum_x));
    return fabs(1.0/(N-1) * (sum_x_squared - (1.0/N) * sum_x*sum_x));
}

double i_mean(const int *array, const int N){
    int i;
    int sum_x;
	if(N==0) return 0;
    sum_x = 0;
    for(i=0; i<N; i++){
        sum_x += array[i];
    }
    return sum_x/(N+0.0);
}
*/

double d_var(const double *array, const int N){
    int i;
    double sum_x, sum_x_squared;
	if(N<=1) return 0.0;
    sum_x = 0;
    sum_x_squared = 0;
    for(i=0; i<N; i++){
        sum_x += array[i];
        sum_x_squared += array[i]*array[i];
    }
    return fabs(1.0/(N-1) * (sum_x_squared - (1.0/N) * sum_x*sum_x));
}

double d_mean(const double *array, const int N){
    int i;
    double sum_x;
	if(N==0) return 0;
    sum_x = 0;
    for(i=0; i<N; i++){
        sum_x += array[i];
    }
    return sum_x/(N+0.0);
}

double d_var_coef(const double *array, const int N){
    int i;
	if(N<=1) return 0;
    return sqrt(d_var(array, N)) / d_mean(array, N);
}

void compute_array_stats(const double *array, const int N, const string array_name, double* results, string* result_names){
//	printf("N in compute_array_stats: %d\n",N);
	result_names[0] = array_name + "__mean";
	results[0] = d_mean(array, N);
	
//	result_names[1] = array_name + "__varcoef";
//	results[1] = d_var_coef(array, N);

//	printf("In compute_array_stats, N=%d\n",N);
//    printMatrixDouble(array, N, 1);
//	printf("%lf \n", array[0]);


	result_names[1] = array_name + "__std";

	results[1] = sqrt(d_var(array, N));

//	cout << result_names[0] << " " << results[0] << "\n";
}

int main (int argc, char *argv[]){
	/* Declare and allocate space for the variables and arrays. */
	set<int>::iterator it;
	vector< set<int> > vset;
	vector< set<int> > variable_sets;
	vector< set<int> > constraint_sets;

	//set<int>::iterator it;
	Stopwatch sw;
	Stopwatch sw2;

	int      solnstat, solnmethod, solntype;
	double   *x     = NULL;
	int      *cstat = NULL;
	int      *rstat = NULL;
	char *xctype;

	CPXENVptr     env = NULL;
	CPXLPptr      lp = NULL;
	int           status = 0;
	int           i,j,k,j2,k2, for_up_bound, count;
	int           cur_numrows, cur_numcols;
	int           method;

	char          *basismsg;

	int *indices = new int[1];
	double *lu = new double[1];
	double *up = new double[1];
	double *lu_eps = new double[1];
	double *up_eps = new double[1];
	lu[0] = 0;
	up[0] = 1;

	char *Lchar = new char[1];
	char *Uchar = new char[1];
	char *Cchar = new char[1];
	Lchar[0] = 'L';
	Uchar[0] = 'U';
	Cchar[0] = 'C';

	int num_var_indices_to_make_cont;
	int *var_indices_to_make_cont;

	double sum = 0;
	double l2 = 0;
	double max_dist = -1;

	int num_binary, num_integer, num_continuous, num_semicont, num_semiint, nzcnt, nzcnt_returned, *rmatbeg, *rmatind, surplus, begin, end, num_nz_this_row;
	double *rmatval, *A_ij_normalized, *a_normalized_varcoefs, a_ik, a_ij_magnitude_mean, a_ij_magnitude_std, a_ij_varcoef_avg, a_ij_varcoef_std, sum_abs;
	double ratio_cont_var;
	int num_non_cont, probtype;
	double mip_obj=hugeValue;

	int n_edges, var_set_num, const_set_num;
	double edge_density;
	/* The following are really integers, but to avoid having to rewrite computation of statistics for ints, we make them doubles. */
	double *vcg_var_degree, *vcg_constraint_degree, *vg_degree, *vg_degree_copy;
	double *vcg_constraint_sum, *vcg_var_sum;
	int **vg_neighbours;
	int vcg_var_deg_min, vcg_var_deg_max, vcg_con_deg_min, vcg_con_deg_max, vg_deg_max, vg_deg_min;
	double vcg_var_deg_mean, vcg_var_deg_std;
	double vcg_con_deg_mean, vcg_con_deg_std;
	double vg_deg_std, vg_deg_q0_25, vg_deg_med, vg_deg_q0_75;
	double *all_clust_coef, clust_coef, deviation, num_neighbour_edges;
	double *obj_fun_coef, *obj_fun_coef_this, *rhs_b, *rhs_b_this, prices_stddev, price_per_number_stddev, price_per_sqrt_number_stddev, *tmp_double_cur_numcols;
	double *up_bound, *low_bound;
	double *support_size;
	int num_stats_of_vector = 2;
	double *tmp_stats = new double[num_stats_of_vector];
	string *tmp_stats_names = new string[num_stats_of_vector];
	int num_q_constr, num_quad, num_qpnz, num_unbounded_discrete;
	double ratio_unbounded_discrete;
	double lp_avg, lp_l2_avg, lp_linf, lp_objval;
	double time_parse, time_nonzero_enum, time_vg, time_clust_coef, time_relax, time_total;
	bool compute_vg = true;
	bool compute_clust_coef = true;
	bool compute_relax = true;
	bool debug = false;
	double *pretime;
	char *sense;
	string var_set_numstr, const_set_numstr;

	vector<string> feature_names(10000);
	vector<double> features(10000);
	int num_features = 0;

	CPXFILEptr fpout  = NULL;
	CPXCHANNELptr cpxresults, cpxwarning, cpxerror, cpxlog;
	char* probfile;
	probfile=new char[256];
	char* probfile2;
	probfile2=new char[256];
	double objval=hugeValue;
	double finalgap=hugeValue;
	int itcnt ;

	sw.Start();
	sw2.Start();

	/* Check the command line arguments */
	if ( argc != 2 ) {
	  usage (argv[0]);
	  goto TERMINATE;
	}

	/* Initialize the CPLEX environment */
	env = CPXopenCPLEX (&status);
	/* If an error occurs, the status value indicates the reason for
	  failure.  A call to CPXgeterrorstring will produce the text of
	  the error message.  Note that CPXopenCPLEX produces no output,
	  so the only way to see the cause of the error is to use
	  CPXgeterrorstring.  For other CPLEX routines, the errors will
	  be seen if the CPX_PARAM_SCRIND indicator is set to CPX_ON.  */
	if ( env == NULL ) {
		char  errmsg[1024];
		fprintf (stderr, "Could not open CPLEX environment.\n");
		CPXgeterrorstring (env, status, errmsg);
		fprintf (stderr, "%s", errmsg);
		goto TERMINATE;
	}

	/* Set node log interval. */
	status = CPXsetintparam (env, CPX_PARAM_MIPINTERVAL, 1);
	if ( status ) {
		fprintf (stderr, 
                 "Failure to set CPX_PARAM_MIPINTERVAL, error %d.\n", status);
		goto TERMINATE;
	}

	/* Set level pf detail for node log. */
	status = CPXsetintparam (env, CPX_PARAM_MIPDISPLAY, 2);
	if ( status ) {
		fprintf (stderr, 
                 "Failure to set CPX_PARAM_MIPDISPLAY, error %d.\n", status);
		goto TERMINATE;
	}

	/* Turn on output to the screen */
	/*   status = CPXsetintparam (env, CPX_PARAM_SCRIND, CPX_ON);
	if ( status ) {
	  fprintf (stderr, 
			   "Failure to turn on screen indicator, error %d.\n", status);
	  goto TERMINATE;
	}*/

	/* Create the problem, using the filename as the problem name */
	lp = CPXcreateprob (env, &status, argv[1]);
	/* A returned pointer of NULL may mean that not enough memory
	  was available or there was some other problem.  In the case of 
	  failure, an error message will have been written to the error 
	  channel from inside CPLEX.  In this example, the setting of
	  the parameter CPX_PARAM_SCRIND causes the error message to
	  appear on stdout.  Note that most CPLEX routines return
	  an error code to indicate the reason for failure.   */
	if ( lp == NULL ) {
	  fprintf (stderr, "Failed to create LP.\n");
	  goto TERMINATE;
	}

    /* Now read the file, and copy the data into the created lp */
	status = CPXreadcopyprob (env, lp, argv[1], NULL);
	if ( status ) {
		fprintf (stderr, "Failed to read and copy the problem data.\n");
		goto TERMINATE;
	}
	time_parse = sw2.Lap();
	if (debug) fprintf(stderr, " parse time = %lfs\n", time_parse);

	/* Get problem type. */
	probtype = CPXgetprobtype (env, lp);
	feature_names[num_features] = "probtype"; features[num_features++] = probtype;

	/* FH TODO: we're outputting basic information here. This information
           should instead be computed and outputted for all variable subsets. */

	/* Ask CPLEX for the problem size.  cur_numrows and cur_numcols store the 
	    current number of rows and columns, respectively.  */
	cur_numcols = CPXgetnumcols (env, lp);
	cur_numrows = CPXgetnumrows (env, lp);
	nzcnt = CPXgetnumnz(env, lp);   /* this does *not* include the nonzero entries for quadratic constraints */
	feature_names[num_features] = "n_vars"; features[num_features++] = cur_numcols;
	feature_names[num_features] = "n_constr"; features[num_features++] = cur_numrows;
	feature_names[num_features] = "n_nzcnt","; features[num_features++] = nzcnt;

	/* Compute features for quadratically-constrained problems */
	num_quad = CPXgetnumquad (env, lp); // number of variables that have quadratic objective coefficients
	num_q_constr = CPXgetnumqconstrs (env, lp);
	num_qpnz = CPXgetnumqpnz (env, lp);
	feature_names[num_features] = "nq_vars","; features[num_features++] = num_quad;
	feature_names[num_features] = "nq_constr","; features[num_features++] = num_q_constr;
	feature_names[num_features] = "nq_nzcnt","; features[num_features++] = num_qpnz;

	/* TODO: here, do all the stuff I can do once I know cur_numcols. */
	/* Get upper and lower bounds. */
	up_bound = new double[cur_numcols];
	low_bound = new double[cur_numcols];
	status = CPXgetub(env, lp, up_bound, 0, cur_numcols-1);
	status = CPXgetlb(env, lp, low_bound, 0, cur_numcols-1);
	
	/* Get variable types. */
	xctype = new char[cur_numcols];
	if (probtype == 0){
		/* LPs don't have ctype information, so manually fill it in. */
		for(i=0; i<cur_numcols; i++){
			xctype[i] = 'C';
		}
	} else {
		status = CPXgetctype (env, lp, xctype, 0, cur_numcols-1);
		if ( status ) {
			fprintf (stderr, "Failed to get the type of variables, error %d.\n", status);
			goto TERMINATE;
		}
	}
	
	/* Collect integer and continuous variables into sets */
	variable_sets = vector< set<int> > (3); /* continuous, integer&binary, all */
	num_binary = 0;
	num_integer = 0;
	num_continuous = 0;
	num_semicont = 0;
	num_semiint = 0;
	
	for(i=0; i<cur_numcols; i++){
		switch ( xctype[i] ) {
			case 'B':
				num_binary++;
				break;
			case 'I':
				num_integer++;
				break;
			case 'C':
				num_continuous++;
				break;
			case 'S':
				num_semicont++;
				break;
			case 'N':
				num_semiint++;
				break;			
		}
		if (xctype[i] == 'C'){ /* not counting semi-continuous variables here! xctype[i] == 'S' */
			variable_sets[1].insert(i);
		} else {
			variable_sets[0].insert(i);
		}
		variable_sets[2].insert(i);
	}
	
	/* Output: basic features */
	feature_names[num_features] = "num_b_variables"; features[num_features++] = num_binary;
	feature_names[num_features] = "num_i_variables"; features[num_features++] = num_integer;
	feature_names[num_features] = "num_c_variables"; features[num_features++] = num_continuous;
	feature_names[num_features] = "num_s_variables"; features[num_features++] = num_semicont;
	feature_names[num_features] = "num_n_variables"; features[num_features++] = num_semiint;
	feature_names[num_features] = "ratio_b_variables"; features[num_features++] = num_binary / (cur_numcols+0.0);
	feature_names[num_features] = "ratio_i_variables"; features[num_features++] = num_integer / (cur_numcols+0.0);
	feature_names[num_features] = "ratio_c_variables"; features[num_features++] = num_continuous / (cur_numcols+0.0);
	feature_names[num_features] = "ratio_s_variables"; features[num_features++] = num_semicont / (cur_numcols+0.0);
	feature_names[num_features] = "ratio_n_variables"; features[num_features++] = num_semiint / (cur_numcols+0.0);
	
	feature_names[num_features] = "num_i+_variables"; features[num_features++] = variable_sets[0].size();
	feature_names[num_features] = "ratio_i+_variables"; features[num_features++] = (variable_sets[0].size() + 0.0) / (cur_numcols+0.0);
	feature_names[num_features] = "n_variables"; features[num_features++] = cur_numcols;
	
	
	/* Compute support sizes for integer variables and percentage of unbounded integer variables */
	support_size = new double[variable_sets[0].size()];
	if (probtype == CPXPROB_LP || probtype == CPXPROB_QP || probtype == CPXPROB_QCP){
		/* These problem types have only continuous variables, so these features are not active (they are conditional!)*/
		num_unbounded_discrete = 0;
	} else {
		/* Percentage continuous or semi-continuous variables, and support sizes for non-continuous  variables. */
		num_non_cont = 0;
		num_unbounded_discrete = 0;
		for (it=variable_sets[0].begin(); it!=variable_sets[0].end(); it++){
			i = *it;
			if (xctype[i] == 'B'){
				support_size[num_non_cont++] = 1+1;
			} else {
				if (xctype[i] == 'I' || xctype[i] == 'N'){
					if (up_bound[i] == CPX_INFBOUND || low_bound[i] == -CPX_INFBOUND){
						num_unbounded_discrete++;
					} 
				}
				
				if (xctype[i] == 'I'){
					support_size[num_non_cont++] = up_bound[i]-low_bound[i]+1;
				} else {
					if (xctype[i] == 'N'){ /* semi-integer */
						support_size[num_non_cont++] = up_bound[i]-low_bound[i]+1+1;
					} else {
						if (xctype[i] == 'S'){ 	/* semi-continuous: either 0 or in that interval, so "2" is the best support size I can come up with */
							support_size[num_non_cont++] = 2;
						} else {
							fprintf (stderr, "WARNING: unknown xctype at position %d: %c. Ignoring this variable in feature computation.\n", i, xctype[i]);
						}
					}
				}	
			}
		}
	}
	feature_names[num_features] = "num_unbounded_disc"; features[num_features++] = num_unbounded_discrete;
	feature_names[num_features] = "ratio_unbounded_disc"; 
	if (variable_sets[0].size() == 0){
		features[num_features++] = 0;
	} else {
		features[num_features++] = (num_unbounded_discrete+0.0)/(variable_sets[0].size());	
	}
	
	/* Compute the stats for the support size vector. */
/*	compute_array_stats(support_size, num_non_cont, "support_size", tmp_stats, tmp_stats_names);
	for (i=0; i<num_stats_of_vector; i++){
		feature_names[num_features] = tmp_stats_names[i];  features[num_features++] = tmp_stats[i];
	}
*/
	obj_fun_coef = new double[cur_numcols];
	sense = new char[cur_numrows];
	rhs_b = new double[cur_numrows];

	/* Get objective function coefficients. */
	status = CPXgetobj (env, lp, obj_fun_coef, 0, cur_numcols-1);
	if ( status ) {
		fprintf (stderr, "Failed to get objective function coefficients. Error %d\n", status);
		goto TERMINATE;
	}
	
	/* Get right hand side, the b array. */
	status = CPXgetrhs (env, lp, rhs_b, 0, cur_numrows-1);
	if ( status ) {
		fprintf (stderr, "Failed to get right hand size, the b array. Error %d\n", status);
		goto TERMINATE;
	}

	
	/* Get sense of constraints. */
	status = CPXgetsense (env, lp, sense, 0, cur_numrows-1);
	if ( status ) {
		fprintf (stderr, "Failed to get constraint sense. Error %d\n", status);
		goto TERMINATE;
	}

	constraint_sets = vector< set<int> > (4); /* 'L', 'E', 'G', 'R' */
	for(i=0; i<cur_numrows; i++){
		switch ( sense[i] ) {
			case 'L':
				constraint_sets[0].insert(i);
				break;
			case 'E':
				constraint_sets[1].insert(i);
				break;
			case 'G':
				constraint_sets[2].insert(i);
				break;
			case 'R':
				constraint_sets[3].insert(i);
				break;			
		}
	}
	
	
/*
	Don't normalize after all, that was only for consistency with Kevin LB's WDP features.
	sum_obj_fun_coef = 0;
	for (i=0; i<cur_numcols; i++){
		sum_obj_fun_coef += obj_fun_coef[i];
	}
	if (sum_obj_fun_coef < 0.0001){ // if sum is too close to zero, don't normalize
		sum_obj_fun_coef = 1;
	}
*/


	for (const_set_num = 0; const_set_num<3; const_set_num++){
		const_set_numstr = convertInt(const_set_num);
		/* Compute features using objective function coefficients. */
		rhs_b_this = new double[constraint_sets[const_set_num].size()];
//		printf("set size: %d\n", constraint_sets[const_set_num].size());
		int num_rhs_b_this = 0;
		for (it=constraint_sets[const_set_num].begin(); it!=constraint_sets[const_set_num].end(); it++){
			i = *it;
//			printf("i=%d, num_rhs_b_this=%d\n", i, num_rhs_b_this);
//			if ( i > 100 ) break;
			rhs_b_this[num_rhs_b_this++] = rhs_b[i];
		}
		/* Compute the stats. */
//		printf("out: %lf \n", rhs_b_this[0]);
		compute_array_stats(rhs_b_this, num_rhs_b_this, "rhs" + var_set_numstr, tmp_stats, tmp_stats_names);
		for (i=0; i<num_stats_of_vector; i++){
			feature_names[num_features] = tmp_stats_names[i];  features[num_features++] = tmp_stats[i];
		}
	}






	/* Allocate memory for results of CPXgetrows. */
	rmatbeg = new int[cur_numrows];
	rmatind = new int[nzcnt];
	rmatval = new double[nzcnt];
	A_ij_normalized = new double[nzcnt];

	/* Get rmatbed, rmatind, and rmatval. */
	status = CPXgetrows(env, lp, &nzcnt_returned, rmatbeg, rmatind, rmatval, nzcnt, &surplus, 0, cur_numrows-1);
	if( nzcnt_returned != nzcnt ){
		printf("nzcnt = %d, nzcnt_returned = %d\n", nzcnt, nzcnt_returned);
		fprintf(stderr, "Error in computation of nonzero entries of the constraint matrix -- value returned by CPXgetrows different than the one from CPXgetnumnz\n" );
		goto TERMINATE;
	}
	if( surplus < 0 ){
		fprintf(stderr, "negative surplus %d, array too small\n", surplus);
		goto TERMINATE;
	} else {
//		fprintf(stderr, "ok surplus %d\n", surplus);
	}


	vcg_constraint_degree = new double[cur_numrows];
	vcg_constraint_sum    = new double[cur_numrows];

	for (var_set_num = 0; var_set_num<3; var_set_num++){
		var_set_numstr = convertInt(var_set_num);

		/* Allocate memory for degree counters. */		
		obj_fun_coef_this = new double[variable_sets[var_set_num].size()];
		
		vcg_var_degree = new double[variable_sets[var_set_num].size()];
		vcg_var_sum    = new double[variable_sets[var_set_num].size()];
		vg_degree = new double[variable_sets[var_set_num].size()];
		vg_degree_copy = new double[variable_sets[var_set_num].size()];
		if (var_set_num == 0){
			vg_neighbours = new double*[variable_sets[var_set_num].size()];
		}

		/* Compute features using objective function coefficients. */
		int num_obj_coefs_this = 0;
		for (it=variable_sets[var_set_num].begin(); it!=variable_sets[var_set_num].end(); it++){
			i = *it;
			obj_fun_coef_this[num_obj_coefs_this++] = obj_fun_coef[i];
		}
		/* Compute the stats. */
		compute_array_stats(obj_fun_coef_this, num_obj_coefs_this, "obj_coefs" + var_set_numstr, tmp_stats, tmp_stats_names);
		for (i=0; i<num_stats_of_vector; i++){
			feature_names[num_features] = tmp_stats_names[i];  features[num_features++] = tmp_stats[i];
		}


		/* Compute variable constraint graph features. */

		/* Initialize  counters. */
		for (i=0; i<cur_numrows; i++) vcg_constraint_degree[i] = 0;
		for (i=0; i<cur_numrows; i++) vcg_constraint_sum[i] = 0;
		for (k=0; k<variable_sets[var_set_num].size(); k++) vcg_var_degree[k] = 0;
		for (k=0; k<variable_sets[var_set_num].size(); k++) vg_degree[k] = 0;
		for (k=0; k<variable_sets[var_set_num].size(); k++) vcg_var_sum[k] = 0;

		vset = vector< set<int> > (variable_sets[var_set_num].size());
		
		/* Enumerate nonzero entries, get variable and constraint degrees, and compute features of A matrix */
		count = 0;
		for(i=0; i<cur_numrows; i++){
			if (i==cur_numrows-1) {
				for_up_bound = nzcnt_returned; 	/* last row, special case for the upper limit of the loop */
			} else {
				for_up_bound = rmatbeg[i+1];
			}
			sum_abs = 0;
			for (j=rmatbeg[i]; j<for_up_bound; j++){
				k = rmatind[j];
				/* If variable k is not an element of the set we're currently considering then skip it. */
				if (variable_sets[var_set_num].count(k) == 0){
					continue;
				}
				
				vcg_constraint_degree[i]++;
				vcg_var_degree[k]++;

				/* For computing features of the rows of the A matrix */
				a_ik = rmatval[j];
//				tmp_double_cur_numcols[j-rmatbeg[i]] = fabs(a_ik); // absolute of nonzero entries in this row.
				sum_abs += fabs(a_ik);
				vcg_var_sum[k] += a_ik;
				vcg_constraint_sum[i] += a_ik;

				if (rhs_b[i] > 1e-6 || rhs_b[i] < -1e-6){
					/* A_ij_normalized holds constraint coefficients of normalized MIP with b = vector of ones (ignoring those rows with b=0 */
					A_ij_normalized[count++] = a_ik / rhs_b[i];
				}

				//printf(" row=%d, j=%d, k=%d, a_ik=%lf, rhs=%lf, begin=%d, end=%d,#vars in constraint=%d\n", i, j, k, a_ik, rhs_b[i], rmatbeg[i], for_up_bound , (for_up_bound - rmatbeg[i] - 1));
			}
		}
		/* Compute the stats for VCG_constr_deg. */
		compute_array_stats(vcg_constraint_degree, cur_numrows, "vcg_constr_deg" + var_set_numstr, tmp_stats, tmp_stats_names);
		for (i=0; i<num_stats_of_vector; i++){
			feature_names[num_features] = tmp_stats_names[i];  features[num_features++] = tmp_stats[i];
		}
		
		/* Compute the stats for VCG_var_deg. */
		compute_array_stats(vcg_var_degree, variable_sets[var_set_num].size(), "vcg_var_deg" + var_set_numstr, tmp_stats, tmp_stats_names);
		for (i=0; i<num_stats_of_vector; i++){
			feature_names[num_features] = tmp_stats_names[i];  features[num_features++] = tmp_stats[i];
		}
		
		/* Compute the stats for VCG_constr_weight. */
		compute_array_stats(vcg_constraint_sum, cur_numrows, "vcg_constr_weight" + var_set_numstr, tmp_stats, tmp_stats_names);
		for (i=0; i<num_stats_of_vector; i++){
			feature_names[num_features] = tmp_stats_names[i];  features[num_features++] = tmp_stats[i];
		}

		/* Compute the stats for VCG_var_weight. */
		compute_array_stats(vcg_var_sum, variable_sets[var_set_num].size(), "vcg_var_weight" + var_set_numstr, tmp_stats, tmp_stats_names);
		for (i=0; i<num_stats_of_vector; i++){
			feature_names[num_features] = tmp_stats_names[i];  features[num_features++] = tmp_stats[i];
		}

/* FH: TODO: not sure where this should go: */
		for(k=0; k<cur_numcols; k++){
			if (vcg_var_degree[k] == 0){
				tmp_double_cur_numcols[k] = obj_fun_coef[k];
			} else {
				tmp_double_cur_numcols[k] = obj_fun_coef[k] / vcg_var_degree[k];
			}
		}
		price_per_number_stddev = sqrt(d_var(tmp_double_cur_numcols, cur_numcols));


		
/* FH Aug 25, 2010: commented, to revisit with Lin 
			// Compute variation coefficient of normalized absolute nonzero entries in this row 
			num_nz_this_row = for_up_bound-rmatbeg[i];
			for (j=0; j<num_nz_this_row; j++){
				tmp_double_cur_numcols[j] /= sum_abs;
			}
			if (num_nz_this_row == 0 ){
				a_normalized_varcoefs[i] = 0;
			} else {
				a_normalized_varcoefs[i] = sqrt(d_var(tmp_double_cur_numcols, num_nz_this_row)) / d_mean(tmp_double_cur_numcols, num_nz_this_row); 
			}
			//printf("num_nz_this_row=%d:\n",num_nz_this_row);
			//printMatrixDouble(tmp_double_cur_numcols, 1, num_nz_this_row);
			
			//printf("rmatbeg:\n");
			//printMatrixInt(rmatbeg, 1, cur_numrows);
*/
			
		
	}

	
	//	price_per_number_stddev = sqrt(d_var(tmp_double_cur_numcols, cur_numcols));

	
	/* Output the computed features. */
	cout << feature_names[0];
	for( i=1; i<num_features; i++ ){
		cout << ", " << feature_names[i];
	}
	cout << "\n";

TERMINATE:
	cout << argv[1] << " ";
	cout << features[0];
	for( i=1; i<num_features; i++ ){
		cout << ", " << features[i];
	}
	
	cout << "\n";
	return (status);
}


void writefeat(string name, double value){
   featname.push_back(name);
   featvalue.push_back(value);
}

void printfeat(){
     vector<string>::iterator i;
     string tmpname;
     vector<double>::iterator j;
     double tmpvalue;
     for (i = featname.begin(); i != featname.end()-1; i++) {
       tmpname=*i;
       cout << tmpname <<", ";
     } 
       tmpname=featname.back();
       cout << tmpname<< endl;
     for (j = featvalue.begin(); j != featvalue.end()-1; j++) {
       tmpvalue=*j;
       cout << tmpvalue <<", ";
     } 
       tmpvalue=featvalue.back();
       cout << tmpvalue<< endl;
      
}

/* This simple routine frees up the pointer *ptr, and sets *ptr to NULL */
static void
free_and_null (char **ptr)
{
   if ( *ptr != NULL ) {
      free (*ptr);
      *ptr = NULL;
   }
} /* END free_and_null */

static void
usage (char *progname)
{
   fprintf (stderr,"Usage: %s filename\n", progname);
   fprintf (stderr,"   where filename is a MIP instance file in .MPS, .LP, or .SAV format.\n");
   fprintf (stderr," Exiting...\n");
} /* END usage */